/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module;

/**
 * @author motohisa
 *
 */
public class Buff<T> {
	
	private String name;
	private int maxSize;
	private T[] qArray;
	private int front;
	private int rear;
	private int nItems;
	
	private boolean isWriteProtected = false;
	private boolean isReadProtected = false;
	
	
	/*
	 * リングバッファ生成
	 * 
	 * @param n バッファ名
	 * @param size バッファサイズ
	 */
	public Buff(String n, int size) {
		
		name = n;
		maxSize = size + 1;
		qArray = (T[]) new Object[maxSize];
		front = 0;
		rear = -1;
		nItems = 0;
		
	}
	
	/*
	 * バッファ書き込み
	 * 
	 * @param o データオブジェクト
	 */
	public synchronized boolean put(T o) {
		
		if (isWriteProtected) {
			return false;
		}
		
		if (nItems >= maxSize) {
			return false;
		}
		if (rear == (maxSize -1)) {
			rear = -1;
		}
		qArray[++rear] = o;
		nItems++;
		
		return true;
	}

	/*
	 * バッファ読み出し
	 * 
	 * @return データオブジェクト
	 */
	public synchronized T get() {
		
		if (isReadProtected) {
			return null;
		}
		
		if (nItems <= 0) {
			return null;
		}
		
		T data = qArray[front++];
		if (front == maxSize) {
			front = 0;
		}
		nItems--;
		
		return data;
	}
	
	/*
	 * バッファの空チェック
	 */
	public boolean isEmpty() {
		return (nItems == 0);
	}

	/*
	 * バッファの満杯チェック
	 */
	public boolean isFull() {
		return (nItems == maxSize);
	}
	
	/*
	 * バッファのサイズ取得
	 */
	public int size() {
		return nItems;
	}

	/*
	 * バッファクリア
	 */
	public synchronized void clear() {
		front = 0;
		rear = -1;
		nItems = 0;	
	}
	
	/*
	 * バッファの書き込みプロテクトの設定
	 */
	public void setWriteProtect(boolean b) {
		isWriteProtected = b;
	}

	/*
	 * バッファの読み出しプロテクトの設定
	 */
	public void setReadProtect(boolean b) {
		isReadProtected = b;
	}

	/*
	 * バッファの書き込みプロテクト状態の取得
	 */
	public boolean isWriteProtected() {
		return isWriteProtected;
	}
	
	/*
	 * バッファ情報表示
	 */
	public void printInfo() {
		System.out.println(" name: " + name + " size: " + maxSize + " fill: " + nItems +
				" RP: " + isReadProtected + " WP: " + isWriteProtected );
	}
	
}
