/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module;

import jp.co.yaskawa.rtc.integratedMotion.controller.CarPosWithElbow;
import jp.co.yaskawa.rtc.integratedMotion.controller.CarPosWithElbowHolder;
import jp.co.yaskawa.rtc.integratedMotion.controller.FrameType;
import jp.co.yaskawa.rtc.integratedMotion.controller.ReturnID;
import jp.co.yaskawa.rtc.integratedMotion.controller.RightLeft;
import jp.co.yaskawa.rtc.integratedMotion.controller.UnitType;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.AbstractMotion;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.CloseGripper;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveCPHoldAbs;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveCPHoldRel;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveGripper;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveJogCPHold;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveJogLinear;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveLinearCartesianAbs;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveLinearCartesianRel;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MovePTPCartesianAbs;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MovePTPCartesianRel;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveUnitAbs;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MoveUnitRel;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.OpenGripper;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Alarm;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Busy;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Off;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Pause;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Powered;
import jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext;
import jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Ready;
import jp.co.yaskawa.rtc.integratedMotion.module.state.StateObjectHolder;
import jp.co.yaskawa.rtc.integratedMotion.module.state.Unpowered;

/**
 * @author motohisa
 *
 */
public class IntegratedMotionInterpolator implements RTEmuContext<IntegratedMotionCommand> {

	final int QUEUE_LENGTH = 8;
	final int NPOSTURE = 24;
	final int NARM_JOINT = 7;
	
	UnitOperation imotionComp_;
	StateObjectHolder<IntegratedMotionCommand> holder_ = new StateObjectHolder<IntegratedMotionCommand>();
	RTEmuUnitState<IntegratedMotionCommand> currentState_;
	Buff<UnitCommand<IntegratedMotionCommand>> commandQ_ = new Buff<UnitCommand<IntegratedMotionCommand>>("command", QUEUE_LENGTH);
	
	AbstractMotion<IntegratedMotionCommand> moveCPHoldAbs_;
	AbstractMotion<IntegratedMotionCommand> moveCPHoldRel_;
	AbstractMotion<IntegratedMotionCommand> moveLinearCarAbs_;
	AbstractMotion<IntegratedMotionCommand> moveLinearCarRel_;
	AbstractMotion<IntegratedMotionCommand> movePTPCarAbs_;
	AbstractMotion<IntegratedMotionCommand> movePTPCarRel_;
	AbstractMotion<IntegratedMotionCommand> moveJogCPHold_;
	AbstractMotion<IntegratedMotionCommand> moveJogLinear_;
	AbstractMotion<IntegratedMotionCommand> moveGripper_;
	AbstractMotion<IntegratedMotionCommand> openGripper_;
	AbstractMotion<IntegratedMotionCommand> closeGripper_;
	AbstractMotion<IntegratedMotionCommand> moveUnitAbs_;
	AbstractMotion<IntegratedMotionCommand> moveUnitRel_;
	AbstractMotion<IntegratedMotionCommand> moveLinearCarAbsRed_;
	AbstractMotion<IntegratedMotionCommand> moveLinearCarRelRed_;
	
	double segmentPeriod_ = 0.01d; //!< セグメント周期[s]
	double[] commandJointPos_ = new double [NARM_JOINT];

	//! 直交指令位置　右アーム x,y,z,R(3x3),左アーム x,y,z,R(3x3)
	double[] commandPosture_ = new double [NPOSTURE];
	boolean[] isConnectUnis_ = new boolean[4];
	int[] axes_info_ = new int[6];
	
	
	public IntegratedMotionInterpolator(UnitOperation u) {
		
		imotionComp_ = u;
		
		// 状態オブジェクトの作成とセット
		holder_.setAlarmStateObject(new Alarm<IntegratedMotionCommand>());
		holder_.setBusyStateObject(new Busy<IntegratedMotionCommand>());
		holder_.setOffStateObject(new Off<IntegratedMotionCommand>());
		holder_.setPauseStateObject(new Pause<IntegratedMotionCommand>());
		holder_.setPoweredStateObject(new Powered<IntegratedMotionCommand>());
		holder_.setReadyStateObject(new Ready<IntegratedMotionCommand>());
		holder_.setUnPoweredStateObject(new Unpowered<IntegratedMotionCommand>());
		
		// 初期状態のセット
		currentState_ = holder_.getOffStateObject();
		
		// コマンドキューのセット
		holder_.getBusyStateObject().setQ(commandQ_);
		holder_.getReadyStateObject().setQ(commandQ_);
		holder_.getAlarmStateObject().setQ(commandQ_);
		
		// モーションオブジェクトのセット
		moveCPHoldAbs_ = new MoveCPHoldAbs<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_CP_HOLD_ABS);
		holder_.getBusyStateObject().putMotion(moveCPHoldAbs_);
		moveCPHoldRel_ = new MoveCPHoldRel<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_CP_HOLD_REL);
		holder_.getBusyStateObject().putMotion(moveCPHoldRel_);
		moveLinearCarAbs_ = new MoveLinearCartesianAbs<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_LINEAR_CAR_ABS);
		holder_.getBusyStateObject().putMotion(moveLinearCarAbs_);
		moveLinearCarRel_ = new MoveLinearCartesianRel<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_LINEAR_CAR_REL);
		holder_.getBusyStateObject().putMotion(moveLinearCarRel_);
		movePTPCarAbs_ = new MovePTPCartesianAbs<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_PTP_CAR_ABS);
		holder_.getBusyStateObject().putMotion(movePTPCarAbs_);
		movePTPCarRel_ = new MovePTPCartesianRel<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_PTP_CAR_REL);
		holder_.getBusyStateObject().putMotion(movePTPCarRel_);
		moveJogCPHold_ = new MoveJogCPHold<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_JOG_CP_HOLD);
		holder_.getBusyStateObject().putMotion(moveJogCPHold_);
		moveJogLinear_ = new MoveJogLinear<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_JOG_LINEAR);
		holder_.getBusyStateObject().putMotion(moveJogLinear_);
		moveGripper_ = new MoveGripper<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_GRIPPER);
		holder_.getBusyStateObject().putMotion(moveGripper_);
		openGripper_ = new OpenGripper<IntegratedMotionCommand>(this,IntegratedMotionCommand.OPEN_GRIPPER);
		holder_.getBusyStateObject().putMotion(openGripper_);
		closeGripper_ = new CloseGripper<IntegratedMotionCommand>(this,IntegratedMotionCommand.CLOSE_GRIPPER);
		holder_.getBusyStateObject().putMotion(closeGripper_);
		moveUnitAbs_ = new MoveUnitAbs<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_UNIT_ABS);
		holder_.getBusyStateObject().putMotion(moveUnitAbs_);
		moveUnitRel_ = new MoveUnitRel<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_UNIT_REL);
		holder_.getBusyStateObject().putMotion(moveUnitRel_);
		moveLinearCarAbsRed_ = new MoveLinearCartesianAbs<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_LINEAR_CAR_ABS_WITH_RED);
		holder_.getBusyStateObject().putMotion(moveLinearCarAbsRed_);
		moveLinearCarRelRed_ = new MoveLinearCartesianRel<IntegratedMotionCommand>(this,IntegratedMotionCommand.MOVE_LINEAR_CAR_REL_WITH_RED);
		holder_.getBusyStateObject().putMotion(moveLinearCarRelRed_);
				
	}
	
	public void changeState(RTEmuUnitState<IntegratedMotionCommand> state) {
		currentState_ = state;
	}
	
	/*
	 * 使用しているコマンドキューを取得する
	 */
	public Buff<UnitCommand<IntegratedMotionCommand>> getQ() {
		return commandQ_;
	}
	
	/*
	 * オン・オフ
	 * @param b 
	 */
	public void setPower(boolean b) {
		currentState_.setPower(this, b);
	}
	
	/*
	 * 主回路電源オン・オフ
	 * @param b
	 */
	public void setAmp(boolean b) {
		currentState_.setAmpPower(this, b);		
	}
	
	/*
	 * サーボオン・オフ
	 * @param b
	 */
	public void setServo(boolean b) {
		currentState_.setServo(this, b);
	}
	
	/*
	 * アラーム状態から復帰する
	 */
	public void clearAlarm() {
		currentState_.clearAlarm(this);
	}
	
	/*
	 * 周期処理をキックする
	 */
	public void doSegment() {
		currentState_.doSegment(this);
	}
	
    /**
     * セグメント周期を設定する
     * @param period 秒単位のセグメント周期
     */
    public void setSegmentPeriod(double period) {
        segmentPeriod_ = period;
    }
    
    /**
     * 設定されているセグメント周期を取得する
     * @return
     */
    public double getSegmentPeriod() {
        return segmentPeriod_;
    }

    /**
     * 直交指令値を取得する
     * 関節角指令値が更新されても、直交指令値には反映されない
     * @return
     */
    double[] getCartesianCommandPos() {
        
        return commandPosture_;
    }

    /**
     * 関節角指令値をセットする
     * @param angle
     */
    public void setJointCommand(UnitType type, double[] angle) {
        int i = 0;
        for(double d : angle) {
        	commandJointPos_[i] = d;
        	i++;
        }
        imotionComp_.setJointCommand(type, angle);
    }    
    
    /**
     * 位置・姿勢データをセットする
     * @param posture 右 x,y,z.HgMatrix(3x3), 左 x,y,z.HgMatrix(3x3)
     */
    public void setCartesianPos(double[] posture ) {
    	int i=0;
    	for(double d: posture) {
    		commandPosture_[i] = d;
    		i++;
    	}
    	imotionComp_.setCartesianCommand(posture);
    }

    public void setRedundantPos(double[] redundantPos ) {
    	imotionComp_.setRedundantCommand(redundantPos);
    }
        
    /**
     * 直交フィードバックをコピーして取得する
     */
    public double[] getCartesianFeedbackPos(RightLeft type, FrameType frameType) {
    	
    	double[] data = new double[13];
    	
    	double[][] mat = new double[3][4];
    	CarPosWithElbow p = new CarPosWithElbow();
    	p.carPos = mat;
    	CarPosWithElbowHolder pos = new CarPosWithElbowHolder(p);
    	
    	imotionComp_.getFeedbackPosCartesian(type, frameType, pos);
    	
    	data[0] = pos.value.carPos[0][3];
    	data[1] = pos.value.carPos[1][3];
    	data[2] = pos.value.carPos[2][3];
    	
    	int index = 3;
		for (int i=0; i<3; i++) {
			for (int j=0; j<3; j++) {
				data[index] = pos.value.carPos[i][j]; //!< HgMat[i][j]
				index++;
			}
		}
		
        data[index] = pos.value.elbow;		

    	return data;
    }
 

    /**
     * 指定ユニットの現在位置（Joint）を取得する
     * @param type
     * @param pos
     * @return
     */
	public int getFeedbackPosJoint(UnitType type, double[] pos) {		
		ReturnID rtn = new ReturnID();
		rtn = imotionComp_.getFeedbackPosJoint(type, pos);

		return rtn.id;	
	}
    
    
    /**
     * 冗長軸の現在位置を取得する
     * @return
     */
    public double[] getRedundantPos() {
    	
    	double[] data = new double[8];
    	
    	imotionComp_.getRedundantPos(data);
    	
    	return data;
    	
    }
    
    public void setConnectUnitsInfo(boolean[] b, int[] axes) {
    	    	
    	for(int i=0; i<isConnectUnis_.length; i++) {
    		isConnectUnis_[i] = b[i];
    	}
    	for(int j=0; j<axes_info_.length; j++) {
    		axes_info_[j] = axes[j];
    	}
    	
    }

	public int getAxesInfo(UnitType type) {		

		if(type.equals(UnitType.RIGHT_ARM)) {
			return axes_info_[0];
		}
		else if(type.equals(UnitType.LEFT_ARM)) {
			return axes_info_[1];
		}
		else if(type.equals(UnitType.LUMBER)) {
			return axes_info_[4];
		}
		else if(type.equals(UnitType.VEHICLE)) {
			return axes_info_[5];
		}
		
		return 0;	
	}
    
    
    /**
     * モーション動作を即停止する
     * @return
     */
    public boolean abort() {
    	return currentState_.abort(this);
    }

    /**
     * モーション動作を一時停止する
     * @return
     */
    public boolean pause() {
    	return currentState_.pause(this);
    }

    /**
     * モーション動作を再開する
     * @return
     */
    public boolean resume() {
    	return currentState_.resume(this);
    }

    /**
     * 状態を取得する
     * @return
     */
    public String getState() {
    	return currentState_.getStateString();
    }
    
    
}
