/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  制御点を維持したまま冗長軸を相対位置指令で動作
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion;

import jp.co.yaskawa.rtc.integratedMotion.controller.FrameType;
import jp.co.yaskawa.rtc.integratedMotion.controller.RightLeft;
import jp.co.yaskawa.rtc.integratedMotion.module.IntegratedMotionInterpolator;
import jp.co.yaskawa.rtc.integratedMotion.module.RTMath;

/**
 * @author motohisa
 *
 */
public class MoveCPHoldRel<T> extends MoveCPHoldAbs<T> {

	public MoveCPHoldRel(IntegratedMotionInterpolator u, T type) {
		super(u, type);
	}

	public void setCommand(UnitCommand<T> c) {

		System.out.println("MoveCPHoldRel: setCommand");
		segmentPeriod_ = unit_.getSegmentPeriod();

		currentRedundantPos_ = unit_.getRedundantPos();
		currentPostureR_ = unit_.getCartesianFeedbackPos(RightLeft.RIGHT, FrameType.ORIGINAL_FRAME);			
		currentPostureL_ = unit_.getCartesianFeedbackPos(RightLeft.LEFT, FrameType.ORIGINAL_FRAME);
		for(int j=0; j<12; j++) {
			currentPosture_[j] = currentPostureR_[j];
			currentPosture_[j+12] = currentPostureL_[j];
		}
				
		double[] params = c.getParams();

		if (params[0] >= 0) {
			rightleft_ = RightLeft.RIGHT; 
			elbow_fb_ = currentRedundantPos_[6];
		} else {
			rightleft_ = RightLeft.LEFT;
			elbow_fb_ = currentRedundantPos_[7];
		}
		lumbarRx_fb_ = currentRedundantPos_[3];
		lumbarRy_fb_ = currentRedundantPos_[4];
		lumbarRz_fb_ = currentRedundantPos_[5];
		vehicleX_fb_ = currentRedundantPos_[0];
		vehicleY_fb_ = currentRedundantPos_[1];
		vehicleR_fb_ = currentRedundantPos_[2];

		elbow_cmd_ = params[1] + elbow_fb_;
		lumbarRx_cmd_ = params[2] + lumbarRx_fb_;
		lumbarRy_cmd_ = params[3] + lumbarRy_fb_;
		lumbarRz_cmd_ = params[4] + lumbarRz_fb_;
		vehicleX_cmd_ = params[5] + vehicleX_fb_;
		vehicleY_cmd_ = params[6] + vehicleY_fb_;
		vehicleR_cmd_ = params[7] + vehicleR_fb_;

    	currentPos_ = new double[7];    	   
        targetPos_ = new double[7];        
		        
        currentPos_[0] = elbow_fb_; 
		currentPos_[1] = lumbarRx_fb_;
		currentPos_[2] = lumbarRy_fb_;
		currentPos_[3] = lumbarRz_fb_;
		currentPos_[4] = vehicleX_fb_;
		currentPos_[5] = vehicleY_fb_;
		currentPos_[6] = vehicleR_fb_;
/*		
		System.out.println("currentPos:");
		for(int i=0; i<currentPos_.length; i++) {
			System.out.println(currentPos_[i]);
		}
*/	
		targetPos_[0] = elbow_cmd_;
		targetPos_[1] = lumbarRx_cmd_;
		targetPos_[2] = lumbarRy_cmd_;
		targetPos_[3] = lumbarRz_cmd_;
		targetPos_[4] = vehicleX_cmd_;
		targetPos_[5] = vehicleY_cmd_;
		targetPos_[6] = vehicleR_cmd_;		
/*
		System.out.println("targetPos:");
		for(int i=0; i<targetPos_.length; i++) {
			System.out.println(targetPos_[i]);
		}
*/			
        deltaPos_ = new double[7];
		absDeltaPos_ = new double[7];
		
		//! 冗長軸の差分値
        deltaPos_[0] = elbow_cmd_ - elbow_fb_; 
        deltaPos_[1] = lumbarRx_cmd_ - lumbarRx_fb_;
        deltaPos_[2] = lumbarRy_cmd_ - lumbarRy_fb_;
        deltaPos_[3] = lumbarRz_cmd_ - lumbarRz_fb_;
        deltaPos_[4] = vehicleX_cmd_ - vehicleX_fb_;
        deltaPos_[5] = vehicleY_cmd_ - vehicleY_fb_;
        deltaPos_[6] = vehicleR_cmd_ - vehicleR_fb_;
/*
		System.out.println("deltaPos:");
		for(int i=0; i<deltaPos_.length; i++) {
			System.out.println(deltaPos_[i]);
		}
*/       
        
        for (int i = 0; i < 7; i++) {
            absDeltaPos_[i] = Math.abs(deltaPos_[i]);
        }
		
		
		demandVelocity_ = new double[6];
		for (int i=0; i<6; i++) {
			demandVelocity_[i] = params[i+8];
		}
/*
		System.out.println("velocity:");
		for(int i=0; i<demandVelocity_.length; i++) {
			System.out.println(demandVelocity_[i]);
		}
*/				
		demandAccel_ = new double[6];
		for (int i=0; i<6; i++) {
			demandAccel_[i] = params[i+14];
		}
/*
		System.out.println("accel:");
		for(int i=0; i<demandAccel_.length; i++) {
			System.out.println(demandAccel_[i]);
		}
*/
		double totalTime[] = new double[7];
		
		for (int i=0; i<totalTime.length; i++) {
			
			if((i==5) || (i==6)){
				if(demandVelocity_[i-1] != 0)
					totalTime[i] = absDeltaPos_[i]/demandVelocity_[i-1];
				else
					totalTime[i] = 0.0;					
			}
			else{
				if(demandVelocity_[i] != 0)
					totalTime[i] = absDeltaPos_[i]/demandVelocity_[i];
				else
					totalTime[i] = 0.0;									
			}	
		}
/*
		System.out.println("time:");
		for(int i=0; i<totalTime.length; i++) {
			System.out.println(totalTime[i]);
		}
*/			
        double maxTime = RTMath.max(totalTime.clone());
        maxAxisNum_ = RTMath.getIndex(maxTime, totalTime);
		
		targetVelocity_ = demandVelocity_[maxAxisNum_];
		accel_ = demandAccel_[maxAxisNum_];

        ratioArray_ = new double[7];
        for (int i = 0; i < ratioArray_.length; i++) {
            ratioArray_[i] = totalTime[i] / totalTime[maxAxisNum_];
            if( deltaPos_[i] <0 )
            {
            	ratioArray_[i] = ratioArray_[i]*(-1);
            }
        }
/*
		System.out.println("ratio:");
		for(int i=0; i<ratioArray_.length; i++) {
			System.out.println(ratioArray_[i]);
		}
*/        
        
		isMoving_ = true;
		
	}
	
}
