/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  関節空間における直線補間（目標位置は直交座標系における絶対位置）
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion;

import jp.co.yaskawa.rtc.integratedMotion.controller.UnitType;
import jp.co.yaskawa.rtc.integratedMotion.module.IntegratedMotionInterpolator;
import jp.co.yaskawa.rtc.integratedMotion.module.Parameters;
import jp.co.yaskawa.rtc.integratedMotion.module.RTMath;
import jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext;

/**
 * @author motohisa
 *
 */
public class MovePTPCartesianAbs<T> extends AbstractMotion<T> {

	IntegratedMotionInterpolator unit_;
	Parameters parameters_;

	double segmentPeriod_ = 0.0d;
    boolean isMoving_ = false;

    int rightArmAxesNum = 0;
    int leftArmAxesNum = 0;
    int lumbarAxesNum = 0;
    int vehicleAxesNum = 0;
        
	double[] currentAngleArray_;
	double[] targetAngleArray_;
	double[] deltaAngleArray_;
	double[] absDeltaAngleArray_;
    double[] ratioArray_;

    double velocity_ = 0.0d;
    double prevVelocity_ = 0.0d;
    double accel_ = 0.0d;
    double targetAngle_ = 0.0d;
    double targetVelocity_ =0.0d;
    int axisNumMaxDistance_;
    	
	public MovePTPCartesianAbs(IntegratedMotionInterpolator u, T type) {
		COMMAND = type;
		unit_ = u;
	}
	
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.AbstractMotion#setParameters(jp.co.yaskawa.rtc.integratedMotion.module.Parameters)
	 */
	@Override
	public void setParameters(Parameters l) {
		parameters_ = l;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) throws MotionAlarmException {
		// TODO Auto-generated method stub
	       isMoving_ = true;
	        
	        velocity_ = velocity_ + accel_ * segmentPeriod_;
	        velocity_ = velocity_ > targetVelocity_ ? targetVelocity_ : velocity_;
	        
	        double remainAngle = Math.abs(targetAngleArray_[axisNumMaxDistance_] - currentAngleArray_[axisNumMaxDistance_]);
	        double velocityReducePoint = Math.sqrt(2 * accel_ * remainAngle) - accel_ * segmentPeriod_ * 0.5d;
	        velocityReducePoint = velocityReducePoint < 0 ? 0 : velocityReducePoint;
	        
	        velocity_ = velocity_ > velocityReducePoint ? velocityReducePoint : velocity_;
	        
	        if (velocity_ * segmentPeriod_ == 0) {
	            isMoving_ = false;
	            currentAngleArray_ = targetAngleArray_;
	        } else {
	            double incrementAngle = 0.0d;
	            for (int i = 0; i < currentAngleArray_.length; i++) {
	                incrementAngle = velocity_ * ratioArray_[i] * segmentPeriod_; //incrementAngle = incrementAngle >= deltaAngleArray_[i] ? deltaAngleArray_[i] : incrementAngle;
	                currentAngleArray_[i] = currentAngleArray_[i] + incrementAngle;
	                deltaAngleArray_[i] = targetAngleArray_[i] - currentAngleArray_[i];
	                absDeltaAngleArray_[i] = Math.abs(deltaAngleArray_[i]);
	            }
	        }
	        
//	        for(int i=0; i<currentAngleArray_.length; i++){
//	        	System.out.println(currentAngleArray_[i]);
//	        }

	        if(rightArmAxesNum>0)
	        {
	    		double[] rArmArray = new double[rightArmAxesNum];
	    		for(int k=0; k<rightArmAxesNum; k++)
	            {
	    			rArmArray[k] = currentAngleArray_[k+6]; 
	            }	
		        unit_.setJointCommand(UnitType.RIGHT_ARM, rArmArray);
	        }
	    	if(leftArmAxesNum >0)
	    	{
	    		double[] lArmArray = new double[leftArmAxesNum];	            
	            for(int k=0; k<leftArmAxesNum; k++)
	            {
	            	lArmArray[k] = currentAngleArray_[k+13]; 
	            }
		        unit_.setJointCommand(UnitType.LEFT_ARM, lArmArray);	            
	    	}
	    	if(lumbarAxesNum >0)
	    	{
	    		double[] lumbarArray = new double[lumbarAxesNum];
	            
	            for(int k=0; k<lumbarAxesNum; k++)
	            {
	            	lumbarArray[k] = currentAngleArray_[k+3]; 
	            }
		        unit_.setJointCommand(UnitType.LUMBER, lumbarArray);	            
	    	}
	        

	        

	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#isMoving()
	 */
	public boolean isMoving() {
		return isMoving_;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.motion.DoMotion#setCommand(jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand)
	 */
	public void setCommand(UnitCommand<T> c) {
		// TODO Auto-generated method stub
    	System.out.println("MovePTPAbs: setCommand");

        int invKinemaDataNum = 24;
        
        segmentPeriod_ = unit_.getSegmentPeriod();

        double[] params = c.getParams();
        rightArmAxesNum = (int)params[0];
        leftArmAxesNum = (int)params[1];
        lumbarAxesNum = (int)params[2];
        vehicleAxesNum = (int)params[3];
                
    	currentAngleArray_ = new double[invKinemaDataNum];
    	if(rightArmAxesNum >0)
    	{
    		double[] rArmArray = new double[rightArmAxesNum];
            unit_.getFeedbackPosJoint(UnitType.RIGHT_ARM, rArmArray);
            
            for(int k=0; k<rightArmAxesNum; k++)
            {
            	currentAngleArray_[k+6] = rArmArray[k]; 
            }
    	}
    	if(leftArmAxesNum >0)
    	{
    		double[] lArmArray = new double[leftArmAxesNum];
            unit_.getFeedbackPosJoint(UnitType.LEFT_ARM, lArmArray);
            
            for(int k=0; k<leftArmAxesNum; k++)
            {
            	currentAngleArray_[k+13] = lArmArray[k]; 
            }
    	}
    	if(lumbarAxesNum >0)
    	{
    		double[] lumbarArray = new double[lumbarAxesNum];
            unit_.getFeedbackPosJoint(UnitType.LUMBER, lumbarArray);
            
            for(int k=0; k<lumbarAxesNum; k++)
            {
            	currentAngleArray_[k+3] = lumbarArray[k]; 
            }
    	}
        
        targetAngleArray_ = new double[invKinemaDataNum];        
        deltaAngleArray_ = new double[invKinemaDataNum];
        absDeltaAngleArray_ = new double[invKinemaDataNum];
        
        for (int i = 0; i < invKinemaDataNum; i++) {
        	targetAngleArray_[i] = params[i+4];
            deltaAngleArray_[i] = targetAngleArray_[i] - currentAngleArray_[i];
            absDeltaAngleArray_[i] = Math.abs(deltaAngleArray_[i]);
        }
        

        double maxAbsoluteAngle = RTMath.max(absDeltaAngleArray_.clone());
        axisNumMaxDistance_ = RTMath.getIndex(maxAbsoluteAngle, absDeltaAngleArray_);
        targetAngle_ = targetAngleArray_[axisNumMaxDistance_];

        targetVelocity_ = maxAbsoluteAngle/params[invKinemaDataNum+4];
        accel_ = targetVelocity_/2;
        
        
        ratioArray_ = new double[invKinemaDataNum];
        for (int i = 0; i < ratioArray_.length; i++) {
            ratioArray_[i] = deltaAngleArray_[i] / absDeltaAngleArray_[axisNumMaxDistance_];
        }
        
        isMoving_ = true;
        
        if (true) {
        	System.out.println("SegmentPeriod: " + segmentPeriod_);
        	System.out.println("Current");
        	System.out.print("\t");
        	for (int i = 0; i < currentAngleArray_.length; i++) {
        		System.out.print(currentAngleArray_[i] + " ");
        	}
        	System.out.println();
        	System.out.println("Target");
        	System.out.print("\t");
        	for (int i = 0; i < targetAngleArray_.length; i++) {
        		System.out.print(targetAngleArray_[i] + " ");
        	}
        	System.out.println();
        	System.out.println("target velocity: " + targetVelocity_ + " accel: " + accel_ );
        	System.out.println("MaxAngle: " + maxAbsoluteAngle + " axis index: " + axisNumMaxDistance_);
        }

	}

}
