/**
 * 
 */
package jp.co.yaskawa.rtc.integratedMotion.module.motion.kinematics;

import javax.vecmath.Matrix4d;

import jp.co.yaskawa.rtc.integratedMotion.module.motion.kinematics.Posture.POSTURE_INDEX;

/**
 * @author motohisa
 *
 */
public class HgTransMatrix extends Matrix4d {
	
    /**
	 * 
	 */
	private static final long serialVersionUID = -488805975011266596L;

	public HgTransMatrix() {
        super(
        		0.0, 0.0, 0.0, 0.0, 
                0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 0.0,
                0.0, 0.0, 0.0, 1.0
                );
    }
    
    public HgTransMatrix(
            double m00, double m01, double m02, double m03,
            double m10, double m11, double m12, double m13,
            double m20, double m21, double m22, double m23) {
        super(
                m00, m01, m02, m03, 
                m10, m11, m12, m13,
                m20, m21, m22, m23,
                0.0, 0.0, 0.0, 1.0
                );
    }
    
    public HgTransMatrix(Posture posture) {
        set(posture);
    }
    
    public HgTransMatrix(HgTransMatrix tmat) {
        super(tmat);
    }
    
    void set(Posture posture) {
        double ca = Math.cos(posture.alpha);
        double sa = Math.sin(posture.alpha);
        double cb = Math.cos(posture.beta);
        double sb = Math.sin(posture.beta);
        double cg = Math.cos(posture.gamma);
        double sg = Math.sin(posture.gamma);
        
        m00 = cb*cg;
        m01 = sa*sb*cg - ca*sg;
        m02 = ca*sb*cg + sa*sg;
        m03 = posture.x;
        
        m10 = cb*sg;
        m11 = sa*sb*sg + ca*cg;
        m12 = ca*sb*sg - sa*cg;
        m13 = posture.y;
        
        m20 = -sb;
        m21 = sa*cb;
        m22 = ca*cb;
        m23 = posture.z;
            
        m30 = m31 = m32 = 0.0;
        m33 = 1.0;
    }

	public double[] toArray() {
	    double[] array = new double[12];
	    array[0] = this.m03;
	    array[1] = this.m13;
	    array[2] = this.m23;
	    array[3] = this.m00;
	    array[4] = this.m01;
	    array[5] = this.m02;
	    array[6] = this.m10;
	    array[7] = this.m11;
	    array[8] = this.m12;
	    array[9] = this.m20;
	    array[10] = this.m21;
	    array[11] = this.m22;	    
	    
	    return array;
	}
    
    

}
