/**

  COPYRIGHT (c) 2010 Yaskawa Electric Corporation.
  
  $Revision: $
  $Author:   $
  $Date:     $
  
  Description:
  アラーム状態クラス
  
 */
package jp.co.yaskawa.rtc.integratedMotion.module.state;

import jp.co.yaskawa.rtc.integratedMotion.module.Buff;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.MotionAlarmException;
import jp.co.yaskawa.rtc.integratedMotion.module.motion.UnitCommand;

/**
 * @author motohisa
 *
 */
public class Alarm<T> implements RTEmuUnitState<T> {

	protected StateObjectHolder<T> holder_;
	protected Buff<UnitCommand<T>> commandQ_;
	private boolean onAlarm_ = false;
	private MotionAlarmException alarm_;
	
	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#abort(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean abort(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#causeAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean causeAlarm(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#clearAlarm(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean clearAlarm(RTEmuContext<T> c) {
		onAlarm_ = false;
		c.changeState(holder_.getReadyStateObject());
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#doSegment(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public void doSegment(RTEmuContext<T> c) {
		commandQ_.clear();
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#pause(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean pause(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#resume(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext)
	 */
	public boolean resume(RTEmuContext<T> c) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setAmpPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setAmpPower(RTEmuContext<T> c, boolean b) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setPower(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setPower(RTEmuContext<T> c, boolean b) {
		if (!b) {
			c.changeState(holder_.getOffStateObject());
			onAlarm_ = false;
		}
		return false;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setServo(jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuContext, boolean)
	 */
	public boolean setServo(RTEmuContext<T> c, boolean b) {
		return true;
	}

	/* (non-Javadoc)
	 * @see jp.co.yaskawa.rtc.integratedMotion.module.state.RTEmuUnitState#setStateObjectHolder(jp.co.yaskawa.rtc.integratedMotion.module.state.StateObjectHolder)
	 */
	public void setStateObjectHolder(StateObjectHolder<T> holder) {
		holder_ = holder;

	}

	/*
	 * コマンドバッファのセット
	 * @param q バッファのオブジェクト
	 */
	public void setQ(Buff<UnitCommand<T>> q){
		commandQ_ = q;
	}

	/*
	 * モーション処理の例外をセットする
	 */
	public void set(MotionAlarmException e) {
		alarm_ = e;
		onAlarm_ = true;
	}

	public String getStateString() {
		return "Alarm";
	}

}
